import json
from collections import defaultdict
from typing import Any, Text, Dict, Union, List

import sqlalchemy as sa
from sqlalchemy.orm import relationship

from rasax.community.legacy.core.domain import (
    Domain as RasaDomain,
    SESSION_CONFIG_KEY,
    SESSION_EXPIRATION_TIME_KEY,
    CARRY_OVER_SLOTS_KEY,
)
from rasax.community.database.base import Base
from rasax.community.database.data import RESPONSE_ANNOTATION_KEYS, Response
from rasax.community.database import utils


def remove_annotations(response_dict: Dict[Text, Any]) -> Dict[Text, Any]:
    """Removes keys from responses that are irrelevant for the dumped domain.

    Args:
        response_dict: Response dictionary to remove the annotation keys from.
        The keys to remove are defined by `RESPONSE_ANNOTATION_KEYS`.

    Returns:
        The response dictionary with removed annotation keys.

    """
    for key in RESPONSE_ANNOTATION_KEYS:
        if key in response_dict:
            del response_dict[key]

    return response_dict


def dump_responses(responses: List[Response]) -> Dict[Text, Any]:
    """Dump responses in a dictionary/list structure.

    Args:
        responses: List of responses to dump.

    Returns:
        Dictionary containing responses as they would be structured in a domain file.
    """
    response_list = [
        {t.response_name: remove_annotations(json.loads(t.content))} for t in responses
    ]

    response_dict = defaultdict(list)
    for response in response_list:
        for k, v in response.items():
            response_dict[k].append(v)

    return dict(response_dict)


class Domain(Base):
    """Stores the domain of the currently deployed Core model."""

    __tablename__ = "domain"

    FILENAME_KEY = "filename"

    id = sa.Column(sa.Integer, utils.create_sequence(__tablename__), primary_key=True)
    project_id = sa.Column(sa.String, sa.ForeignKey("project.project_id"))
    store_entities_as_slots = sa.Column(sa.Boolean, default=True)
    filename = sa.Column(sa.String, nullable=False)
    session_expiration_time = sa.Column(sa.Float)
    carry_over_slots = sa.Column(sa.Boolean)
    to_delete = sa.Column(sa.Boolean, nullable=False, default=False)

    actions = relationship(
        "DomainAction",
        cascade="all, delete-orphan",
        back_populates="domain",
        order_by=lambda: DomainAction.action_id.asc(),
    )
    intents = relationship(
        "DomainIntent",
        cascade="all, delete-orphan",
        back_populates="domain",
        order_by=lambda: DomainIntent.domain_id.asc(),
    )
    entities = relationship(
        "DomainEntity",
        cascade="all, delete-orphan",
        back_populates="domain",
        order_by=lambda: DomainEntity.entity_id.asc(),
    )
    slots = relationship(
        "DomainSlot",
        cascade="all, delete-orphan",
        back_populates="domain",
        order_by=lambda: DomainSlot.slot_id.asc(),
    )
    responses = relationship(
        "Response",
        cascade="all, delete-orphan",
        back_populates="domain",
        order_by=lambda: Response.id.asc(),
    )

    def _get_session_config(self) -> Dict[Text, Union[bool, float]]:
        session_config = {}

        if self.session_expiration_time is not None:
            # use an `int` if the `float` value is actually an `int`
            session_config[SESSION_EXPIRATION_TIME_KEY] = (
                int(self.session_expiration_time)
                if self.session_expiration_time == int(self.session_expiration_time)
                else self.session_expiration_time
            )

        if self.carry_over_slots is not None:
            session_config[CARRY_OVER_SLOTS_KEY] = self.carry_over_slots

        return session_config

    def as_dict(self) -> Dict[Text, Any]:
        """Returns domain in format which can be serialized."""
        domain_dict = {
            "config": {"store_entities_as_slots": self.store_entities_as_slots},
            "actions": [e.action for e in self.actions if not e.is_form],
            "forms": {e.action: e.as_dict()["data"] for e in self.actions if e.is_form},
            "entities": [e.entity for e in self.entities],
            "intents": [i.as_dict() for i in self.intents],
            "slots": {s.slot: s.as_dict() for s in self.slots},
            "responses": dump_responses(self.responses),
        }

        if self.filename:
            domain_dict[self.FILENAME_KEY] = self.filename.lstrip("./")

        session_config = self._get_session_config()
        if session_config:
            domain_dict[SESSION_CONFIG_KEY] = session_config

        return domain_dict

    def as_rasa_domain(self) -> RasaDomain:
        """Returns Domain as Rasa Domain object."""
        return RasaDomain.from_dict(self.as_dict())

    def is_empty(self):
        """Return true if it contains to domain-specific objects."""
        return not any(
            [self.actions, self.responses, self.entities, self.intents, self.slots]
        )


class DomainAction(Base):
    """Stores the actions which are defined in the domain."""

    __tablename__ = "domain_action"

    action_id = sa.Column(
        sa.Integer, utils.create_sequence(__tablename__), primary_key=True
    )
    domain_id = sa.Column(sa.Integer, sa.ForeignKey("domain.id"))
    action = sa.Column(sa.String)
    is_form = sa.Column(sa.Boolean, default=False)

    # Starting from Rasa Open Source 2.0, slot mappings for forms are defined
    # in the domain file. To store these mappings, we use the `data` column to
    # store the JSON-serialized information there.
    data = sa.Column(sa.Text)

    domain = relationship("Domain", back_populates="actions")

    def as_dict(self) -> Dict[Text, Any]:
        """Return Action as dictionary."""
        return {
            "id": self.action_id,
            "domain_id": self.domain_id,
            "name": self.action,
            "is_form": self.is_form,
            "data": json.loads(self.data) if self.data else {},
        }


class DomainIntent(Base):
    """Stores the intents which are defined in the domain."""

    __tablename__ = "domain_intent"

    intent_id = sa.Column(
        sa.Integer, utils.create_sequence(__tablename__), primary_key=True
    )
    domain_id = sa.Column(sa.Integer, sa.ForeignKey("domain.id"))
    intent = sa.Column(sa.String)
    triggered_action = sa.Column(sa.String)
    use_entities = sa.Column(sa.Text)
    ignore_entities = sa.Column(sa.Text)

    domain = relationship("Domain", back_populates="intents")

    def as_dict(self) -> Dict[Text, Any]:
        """Return Intent as dictionary."""
        config = {
            "use_entities": json.loads(self.use_entities or "true"),
            "ignore_entities": json.loads(self.ignore_entities or str([])),
        }

        if self.triggered_action:
            config["triggers"] = self.triggered_action

        return {self.intent: config}


class DomainEntity(Base):
    """Stores the entities which are defined in the domain."""

    __tablename__ = "domain_entity"

    entity_id = sa.Column(
        sa.Integer, utils.create_sequence(__tablename__), primary_key=True
    )
    domain_id = sa.Column(sa.Integer, sa.ForeignKey("domain.id"))
    entity = sa.Column(sa.String)

    domain = relationship("Domain", back_populates="entities")


class DomainSlot(Base):
    """Stores the slots which are defined in the domain."""

    __tablename__ = "domain_slot"

    slot_id = sa.Column(
        sa.Integer, utils.create_sequence(__tablename__), primary_key=True
    )
    domain_id = sa.Column(sa.Integer, sa.ForeignKey("domain.id"))
    slot = sa.Column(sa.String)
    mappings = sa.Column(sa.Text, nullable=False, server_default=json.dumps([]))
    initial_value = sa.Column(sa.Text)
    type = sa.Column(
        sa.String, default="rasa.shared.core.slots.AnySlot"
    )  # TODO(alwx): add a migration?
    values = sa.Column(sa.Text)
    max_value = sa.Column(sa.Float)
    min_value = sa.Column(sa.Float)
    influence_conversation = sa.Column(sa.Boolean)
    custom_keys = sa.Column(sa.Text)

    domain = relationship("Domain", back_populates="slots")

    def as_dict(self) -> Dict[Text, Any]:
        """Returns slots in format which can be serialized."""
        slot = {}

        if self.custom_keys:
            slot = json.loads(self.custom_keys)

        slot["type"] = self.type
        slot["mappings"] = json.loads(self.mappings)

        if self.initial_value:
            slot["initial_value"] = json.loads(self.initial_value)

        if "categorical" in self.type.lower() and self.values:
            slot["values"] = json.loads(self.values)

        for prop in ["max_value", "min_value", "influence_conversation"]:
            # Only include these properties in the slot dictionary if they
            # have been explicitly set.
            value = getattr(self, prop)
            if value is not None:
                slot[prop] = value

        return slot
