import logging
from typing import Dict, List, Optional, Text, Union

import rasax.community.legacy.data
from rasax.community.legacy.core.training_data.story_reader.yaml_story_reader import (
    YAMLStoryReader,
)
from rasax.community.legacy.core.training_data.structures import StoryGraph
from rasax.community.legacy.importers import utils
from rasax.community.legacy.importers import autoconfig
from rasax.community.legacy.importers.importer import TrainingDataImporter
from rasax.community.legacy.importers.autoconfig import TrainingType
from rasax.community.legacy.nlu.training_data.training_data import TrainingData
from rasax.community.legacy.core.domain import InvalidDomain, Domain
import rasax.community.legacy.utils.io

logger = logging.getLogger(__name__)


class RasaFileImporter(TrainingDataImporter):
    """Default `TrainingFileImporter` implementation."""

    def __init__(
        self,
        config_file: Optional[Text] = None,
        domain_path: Optional[Text] = None,
        training_data_paths: Optional[Union[List[Text], Text]] = None,
        training_type: Optional[TrainingType] = TrainingType.BOTH,
    ):

        self._domain_path = domain_path

        self._nlu_files = rasax.community.legacy.data.get_data_files(
            training_data_paths, rasax.community.legacy.data.is_nlu_file
        )
        self._story_files = rasax.community.legacy.data.get_data_files(
            training_data_paths, YAMLStoryReader.is_stories_file
        )
        self._conversation_test_files = rasax.community.legacy.data.get_data_files(
            training_data_paths, YAMLStoryReader.is_test_stories_file
        )

        self.config = autoconfig.get_configuration(config_file, training_type)

    def get_config(self) -> Dict:
        """Retrieves model config (see parent class for full docstring)."""
        return self.config

    def get_stories(self, exclusion_percentage: Optional[int] = None) -> StoryGraph:
        """Retrieves training stories / rules (see parent class for full docstring)."""
        return utils.story_graph_from_paths(
            self._story_files, self.get_domain(), exclusion_percentage
        )

    def get_conversation_tests(self) -> StoryGraph:
        """Retrieves conversation test stories (see parent class for full docstring)."""
        return utils.story_graph_from_paths(
            self._conversation_test_files, self.get_domain()
        )

    def get_nlu_data(self, language: Optional[Text] = "en") -> TrainingData:
        """Retrieves NLU training data (see parent class for full docstring)."""
        return utils.training_data_from_paths(self._nlu_files, language)

    def get_domain(self) -> Domain:
        """Retrieves model domain (see parent class for full docstring)."""
        domain = Domain.empty()

        # If domain path is None, return an empty domain
        if not self._domain_path:
            return domain
        try:
            domain = Domain.load(self._domain_path)
        except InvalidDomain as e:
            rasax.community.legacy.utils.io.raise_warning(
                f"Loading domain from '{self._domain_path}' failed. Using "
                f"empty domain. Error: '{e}'"
            )

        return domain
